#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#include <math.h>
#include <float.h>
#include "klrversion.h"
#include "globals.h"
#include "example_set.h"
#include "parameters.h"
#include "kernel.h"

// for timing
//#include <time.h>
//#include <sys/types.h>
//#include <unistd.h>


// global kernel-objects
kernel_c* kernel=0;
parameters_c* parameters=0;
example_set_c* training_set=0;


// examples
int n; // #examples
double* target;
long N1;
long N2;

// internal vars
double* alpha;
double* Hcache;
int* at_bound;
long i_up;
long i_low;
double b;
double b_up;
double b_low;


// user-defined params
double tol; // Genauigkeit fr b (convergence_epsilon)
double C;
double epsilon; // Genauigkeit fr t (is_zero)
double mu; // Genauigkeit fr bound (is_zero)
long max_iterations; // maximale Anzahl Iterationen im Newton-Raphson-Schritt

double loginfinity = log(infinity);



inline
double dG(double alpha){
  // return dG(alpha/C)
  return log(alpha/(C-alpha));
};


inline
double dPhi(double t,long i, long j,double ai, double aj, double Kii, double Kij, double Kjj){
  // return Hi(alpha(t)) - Hj(alpha(t))
  double result=0.0;
  double ydG=0.0;
  if(target[i] > 0){
    result = Kii-Kij;
    ydG = dG(ai+t)-dG(ai);
  }
  else{
    result = Kij-Kii;
    ydG = dG(ai)-dG(ai-t);
  };
  if(target[j] > 0){
    result = Kjj-Kij;
    ydG -= dG(aj-t)-dG(aj);
  }
  else{
    result = Kij-Kjj;
    ydG -= dG(aj)-dG(aj+t);
  };
  //  result *= t;
  result = t*(Kii-2.0*Kij+Kjj);
  result += ydG;
  result += Hcache[i]-Hcache[j]; // value for t=0

  return result;
};


inline
double d2Phi(double t,long i, long j,double ai, double aj, double Kii, double Kij, double Kjj){
  double result;
  double atilde;
  if(target[i] > 0){
    atilde = ai+t;
  }
  else{
    atilde = ai-t;
  };
  result = C/((atilde)*(C-atilde));
  if(target[j] > 0){
    atilde = aj-t;
  }
  else{
    atilde = aj+t;
  };
  result += C/((atilde)*(C-atilde));

  result+=Kii-2.0*Kij+Kjj; // eta

  return result;
};


int takeStep (long i, long j){
  double* kernel_row_i = kernel->get_row(i);
  double* kernel_row_j = kernel->get_row(j);
  double aio = alpha[i];
  double ajo = alpha[j];
  double yi = target[i];
  double yj = target[j];
  double Hio = Hcache[i];
  double Hjo = Hcache[j];
  double Kii = kernel_row_i[i];
  double Kij = kernel_row_i[j];
  double Kjj = kernel_row_j[j];
  int takestepFlag = 1;

  // Compute t_min and t_max 
  double t_max;
  double t_min;
  double t_tmp;
  if(yi > 0){
    t_min = mu/2.0-aio;
    t_max = C-mu/2.0-aio;
  }
  else{
    t_max = aio-mu/2.0;
    t_min = aio-(C-mu/2.0);
  };
  if(yj > 0){
    t_tmp = ajo-mu/2.0;
    if(t_tmp < t_max){ t_max = t_tmp; };
    t_tmp = ajo-(C-mu/2.0);
    if(t_tmp > t_min){ t_min = t_tmp; };
  }
  else{
    t_tmp = mu/2.0-ajo;
    if(t_tmp > t_min){ t_min = t_tmp; };
    t_tmp = C-mu/2.0-ajo;
    if(t_tmp < t_max){ t_max = t_tmp; };
  };

  if (t_max - t_min <= epsilon){
    return 0;
  }; 
  double t = 0.0;
  double the_dPhi = Hio - Hjo;
  double the_d2Phi = Kii - 2.0*Kij + Kjj + C/(aio * (C - aio)) + C/(ajo * (C - ajo));
  double dPhi_left;
  double d2Phi_left;
  double dPhi_right;
  double d2Phi_right;
  double t_left;
  double t_right;

  if (the_dPhi > 0) { 
    //Compute dPhi(t) and d2Phi(t) at t = t_min and denoted as dPhi_left, d2Phi_left 
    dPhi_left = dPhi(t_min,i,j,aio,ajo,Kii,Kij,Kjj);
    d2Phi_left = d2Phi(t_min,i,j,aio,ajo,Kii,Kij,Kjj);
    if (dPhi_left < 0) { 
      t_left = t_min;
      t_right = t;
      dPhi_right = the_dPhi;
      d2Phi_right = the_d2Phi;
    } 
    else { 
      t = t_min;
      takestepFlag = 2;
    };
  } 
  else if (the_dPhi < 0) {
    //Compute dPhi(t) and d2Phi(t) at t = t_max and denoted as dPhi_right, d2Phi_right 
    dPhi_right = dPhi(t_max,i,j,aio,ajo,Kii,Kij,Kjj);
    d2Phi_right = d2Phi(t_max,i,j,aio,ajo,Kii,Kij,Kjj); 
    if (dPhi_right > 0) { 
      t_left = t;
      t_right = t_max;
      dPhi_left = the_dPhi;
      d2Phi_left = the_d2Phi;
    } 
    else { 
      t = t_max;
      takestepFlag = 2;
    };
  }
  else { 
    return 0;
  };

  double t0;
  double dt;
  double ai;
  double aj;
  double Hi;
  double Hj;
  if (takestepFlag == 1) { 
    //Choose a better start point
    if (abs(dPhi_left) < abs(dPhi_right)) { 
      t0 = t_left;
      the_dPhi = dPhi_left;
      the_d2Phi = d2Phi_left;
    } 
    else { 
      t0 = t_right;
      the_dPhi = dPhi_right;
      the_d2Phi = d2Phi_right;
    }
    do{ 
      dt = - the_dPhi / the_d2Phi; // Newton-Raphson step 
      t = t0 + dt;
      if((t <= t_left) || (t >= t_right)){ 
	t = (t_left + t_right) / 2.0; // Bisection step
      };

      ai = aio + t/yi;
      aj = ajo - t/yj;

      Hi = Hio + t*(Kii - Kij) + yi * (log(ai/(C - ai))-log(aio/(C-aio)));
      Hj = Hjo + t*(Kij - Kjj) + yj * (log(aj/(C - aj))-log(ajo/(C-ajo)));

      the_dPhi = Hi - Hj;
      the_d2Phi = Kii - 2*Kij + Kjj + C/(ai * (C - ai)) + C/(aj * (C - aj));

      //Update t_left and t_right 
      if (the_dPhi * dPhi_left > 0){
	t_left = t;
	dPhi_left = the_dPhi;
      }
      else{ 
	t_right = t;
	dPhi_right = the_dPhi;
      };
      t0 = t;
    } while((abs(the_dPhi) > (0.1 * tol)) 
	    && (t_left + epsilon < t_right));
  }
  else if (takestepFlag == 2){
    ai = aio + t / yi;
    aj = ajo - t / yj;
    Hi = Hio + t*(Kii - Kij) + yi * (log(ai/(C - ai))-log(aio/(C-aio)));
    Hj = Hjo + t*(Kij - Kjj) + yj * (log(aj/(C - aj))-log(ajo/(C-ajo)));
  };

  if (t == 0){ 
    return 0; 
  };

  // Save ai, aj, Hi, Hj 
  alpha[i] = ai;
  alpha[j] = aj;
  // Update the boundary property of indices i and j (evt. clip alpha)
  if(ai <= mu){
    if(((target[i] > 0) && (target[j] > 0)) ||
       ((target[i] < 0) && (target[j] < 0))){
      alpha[j] -= (mu-alpha[i]);
    }
    else{
      alpha[j] += (mu-alpha[i]);
    };
    alpha[i] = mu;
    at_bound[i]=1;
  }
  else if(ai >= C-mu){
    if(((target[i] > 0) && (target[j] > 0)) ||
       ((target[i] < 0) && (target[j] < 0))){
      alpha[j] -= (C-mu-alpha[i]);
    }
    else{
      alpha[j] += (C-mu-alpha[i]);
    };
    alpha[i] = C-mu;
    at_bound[i]=1;
  }
  else{
    at_bound[i]=0;
  };
  if(aj <= mu){
    if(((target[i] > 0) && (target[j] > 0)) ||
       ((target[i] < 0) && (target[j] < 0))){
      alpha[i] -= (mu-alpha[j]);
    }
    else{
      alpha[i] += (mu-alpha[j]);
    };
    alpha[j] = mu;
    at_bound[j]=1;
  }
  else if(aj >= C-mu){
    if(((target[i] > 0) && (target[j] > 0)) ||
       ((target[i] < 0) && (target[j] < 0))){
      alpha[i] -= (C-mu-alpha[j]);
    }
    else{
      alpha[i] += (C-mu-alpha[j]);
    };
    alpha[j] = C-mu;
    at_bound[j]=1;
  }
  else{
    at_bound[j]=0;
  };

  t = ((alpha[i]-aio)*yi+(ajo-alpha[j])*yj)/2.0;

  Hi = Hio + t*(Kii - Kij) + yi * (log(alpha[i]/(C - alpha[i]))-log(aio/(C - aio)));
  Hj = Hjo + t*(Kij - Kjj) + yj * (log(alpha[j]/(C - alpha[j]))-log(ajo/(C - ajo)));

  for(int k=0;k<n;k++){
    Hcache[k] += t*(kernel_row_i[k]-kernel_row_j[k]);
  };

  Hcache[i] = Hi;
  Hcache[j] = Hj;

  // Update i_low, i_up, b_low and b_up over indices of non-boundary group 
  b_up = -infinity;
  b_low = infinity;
  i_up = 0;
  i_low = 0;
  for(long i=0;i<n;i++){
    if(! at_bound[i]){
      if(Hcache[i] > b_up){
	b_up = Hcache[i];
	i_up = i;
      };
      if(Hcache[i] < b_low){
	b_low = Hcache[i];
	i_low = i;
      };
    };
  };

  return 1;
}; // end takeStep procedure


void klr(){
  // main routine: 

  // precond: n, target, examples, kernel intialisiert
  alpha = new double[n];
  Hcache = new double[n];
  at_bound = new int[n];

  long i;
  long j;
  double alpha_pos = C/N1; // !!! N1 = # pos examples
  double alpha_neg = C/N2; // !!! N2 = # neg examples
  for(i=0;i<n;i++){
    if(target[i] > 0){
      alpha[i] = alpha_pos;
    }
    else{
      alpha[i] = alpha_neg;
    };
    at_bound[i] = 0;
  };

  // initialize all Hcache[i]
  double sum_pos_K;
  double sum_neg_K;
  double* kernel_row;
  b_up = -infinity;
  b_low = infinity;
  i_up = 0;
  i_low = 0;
  for(i=0;i<n;i++){
    sum_pos_K = 0.0;
    sum_neg_K = 0.0;
    kernel_row = kernel->get_row(i);
    for(j=0;j<n;j++){
      if(target[j] > 0){
	sum_pos_K += kernel_row[j];
      }
      else{
	sum_neg_K += kernel_row[j];
      };
    };
    Hcache[i] = alpha_pos*sum_pos_K-alpha_neg*sum_neg_K
      +(double)(target[i])*dG(alpha[i]);
    if(Hcache[i] > b_up){
      b_up = Hcache[i];
      i_up = i;
    };
    if(Hcache[i] < b_low){
      b_low = Hcache[i];
      i_low = i;
    };
  };

  int Flag;
  double Hi;
  long numChange;

  long it=max_iterations;

  do{ 
    //takestep with i_low and i_up
    while (2.0*tol < b_up - b_low) { 
      it--;

      //     cout<<"."; cout.flush();
      Flag = takeStep(i_low, i_up);
      if (Flag == 0) { break; };
    };
    //    cout<<endl;

    //check optimality of boundary indices 
    numChange = 0;
    for (i = 0; i < n; i++){ 
      if (at_bound[i]) { 
	Hi = Hcache[i];
	if ( abs(Hi - b_low) >= abs(Hi - b_up)) { 
	  it--;
	  Flag = takeStep(i, i_low);
	  if (Flag == 0) { 
	    it--;
	    Flag = takeStep(i, i_up);
	  };
	} 
	else { 
	  it--;
	  Flag = takeStep(i, i_up);
	  if (Flag == 0) { 
	    it--;
	    Flag = takeStep (i, i_low);
	  };
	};
	if(! at_bound[i]){
	  numChange++;
	};
      };
    };
  } while((numChange != 0) && (it > 0));

  cout<<"...done, KKT error = "<<((b_up - b_low)/2.0)<<endl;

  cout<<(max_iterations-it)<<" iterations"<<endl;

  b = (b_low+b_up)/2.0;

//    double targetf = 0.0;
//    double tmp;
//    for(long i=0;i<n;i++){ 
//      double* kernel_row_i = kernel->get_row(i);
//      for(long j=0;j<n;j++){
//        targetf += 0.5*alpha[i]*alpha[j]
//  	*target[i]*target[j]
//  	*kernel_row_i[j];
//      };
//      tmp = alpha[i]/C;
//      targetf += C*(tmp*log(tmp)+(1-tmp)*log(1-tmp));
//    };
//    cout<<"TARGET = "<<targetf<<endl;

}; //end main rountine 


/////////////////////////////////////////////////////////////////////////

struct example_set_list{
  example_set_c* the_set;
  example_set_list* next;
};
example_set_list* test_sets = 0;


void print_help(){
  cout<<endl;
  cout<<"myklr: do kernel logistic regression from the given parameters and examples."<<endl<<endl;
  cout<<"usage: myklr"<<endl
      <<"       myklr <FILE>"<<endl
      <<"       myklr <FILE1> <FILE2> ..."<<endl<<endl;
  cout<<"The input has to consist of:"<<endl
      <<"- the klr parameters"<<endl
      <<"- the kernel definition"<<endl
      <<"- the training set"<<endl
      <<"- one or more test sets (optional)"<<endl;

  cout<<endl<<"See the documentation for the input format. The first example set to be entered is considered to be the training set, all others are test sets. Each input file can consist of one or more definitions. If no input file is specified, the input is read from <stdin>."<<endl<<endl;

  cout<<endl<<"This software is free only for non-commercial use. It must not be modified and distributed without prior permission of the author. The author is not responsible for implications from the use of this software."<<endl;
  exit(0);
};


void read_input(istream& input_stream, char* filename){
  // returns number of examples sets read
  char* s = new char[MAXCHAR];
  char next;
  int is_linear=1; // linear kernel?
  next = input_stream.peek();
  if(next == EOF){ 
    // set stream to eof
    next = input_stream.get(); 
  };
  while(! input_stream.eof()){
    if('#' == next){
      // ignore comment
      input_stream.getline(s,MAXCHAR);
    }
    else if(('\n' == next) ||
	    (' ' == next) ||
	    ('\r' == next) ||
	    ('\f' == next) ||
	    ('\t' == next)){
      // ignore
      next = input_stream.get();
    }
    else if('@' == next){
      // new section
      input_stream >> s;
      if(0 == strcmp("@parameters",s)){
	// read parameters
	if(parameters == 0){
	  parameters = new parameters_c();
	  input_stream >> *parameters;
	  if(kernel != 0){
	    parameters->is_linear = is_linear;
	  };
	}
	else{
	  cout <<"*** ERROR: Parameters multiply defined"<<endl;
	  throw input_exception();
	};
      }
      else if(0==strcmp("@examples",s)){
	if(0 == training_set){
	  // input training set
	  training_set = new example_set_c();
	  if(0 != parameters){
	    training_set->set_format(parameters->default_example_format);
	  };
	  input_stream >> *training_set;	    
	  training_set->set_filename(filename);
	  cout<<"   read "<<training_set->size()<<" examples, format "<<training_set->my_format<<", dimension = "<<training_set->get_dim()<<"."<<endl;
	}
	else{
	  // input test sets
	  example_set_list* test_set = new example_set_list;
	  test_set->the_set = new example_set_c();
	  if(0 != parameters){
	    (test_set->the_set)->set_format(parameters->default_example_format);
	  };
	  input_stream >> *(test_set->the_set);
	  (test_set->the_set)->set_filename(filename);
	  test_set->next = test_sets;
	  test_sets = test_set;
	  cout<<"   read "<<(test_set->the_set)->size()<<" examples, format "<<(test_set->the_set)->my_format<<", dimension = "<<(test_set->the_set)->get_dim()<<"."<<endl;
	};
      }
      else if(0==strcmp("@kernel",s)){
	if(0 == kernel){
	  kernel_container_c k_cont;
	  input_stream >> k_cont;
	  kernel = k_cont.get_kernel();
	  is_linear = k_cont.is_linear;
	  if(parameters != 0){
	    parameters->is_linear = is_linear;
	  };
	}
	else{
	  cout <<"*** ERROR: Kernel multiply defined"<<endl;
	  throw input_exception();
	};
      };
    }
    else{
      // default = "@examples"
      if(0 == training_set){
	// input training set
	training_set = new example_set_c();
	if(0 != parameters){
	  training_set->set_format(parameters->default_example_format);
	};
	input_stream  >> *training_set;	    
	training_set->set_filename(filename);
	cout<<"   read "<<training_set->size()<<" examples, format "<<training_set->my_format<<", dimension = "<<training_set->get_dim()<<"."<<endl;
      }
      else{
	// input test sets
	example_set_list* test_set = new example_set_list;
	test_set->the_set = new example_set_c();
	if(0 != parameters){
	  (test_set->the_set)->set_format(parameters->default_example_format);
	};
	input_stream >> *(test_set->the_set);
	(test_set->the_set)->set_filename(filename);
	test_set->next = test_sets;
	test_sets = test_set;
	cout<<"   read "<<(test_set->the_set)->size()<<" examples, format "<<(test_set->the_set)->my_format<<", dimension = "<<(test_set->the_set)->get_dim()<<"."<<endl;
      };
    };
    next = input_stream.peek();
    if(next == EOF){ 
      // set stream to eof
      next = input_stream.get(); 
    };
  };
  delete []s;
};


double predict(svm_example example){
  long i;
  svm_example sv;
  double the_sum=b;

  for(i=0;i<n;i++){
    if(alpha[i] != 0){
      sv = training_set->get_example(i);
      the_sum += alpha[i]*kernel->calculate_K(sv,example);
    };
  };
  return the_sum;
};


void predict(example_set_c* test_set){
  long i;
  double prediction;
  svm_example example;

  for(i=0;i<test_set->size();i++){
    example = test_set->get_example(i);
    prediction = predict(example);

//      prediction = -prediction;
//      if(prediction > loginfinity){
//        prediction = 0;
//      }
//      else{
//        prediction = 1.0/(1.0+exp(prediction));
//      };

    test_set->put_y(i,prediction);
  };
  test_set->set_initialised_y();
  test_set->put_b(b);
  if(parameters->verbosity>=4){
    cout<<"Prediction generated"<<endl;
  };
};


/////////////////////////////////////////////////////////////////////////


int main(int argc,char* argv[]){
  cout<<"*** myKLR version "<<klrversion<<" ***"<<endl;
  cout.precision(8);
  // read objects
  try{
    if(argc<2){
      cout<<"Reading from STDIN"<<endl;
      // read vom cin
      read_input(cin,"klr");
    }
    else{
      char* s = argv[1];
      if((0 == strcmp("-h",s)) || (0==strcmp("-help",s)) || (0==strcmp("--help",s))){
	// print out command-line help
	print_help();
      }
      else{
	// read in all input files
	for(int i=1;i<argc;i++){
	  if(0 == strcmp(argv[i],"-")){
	    cout<<"Reading from STDIN"<<endl;
	    // read vom cin
	    read_input(cin,"klr");
	  }
	  else{
	    cout<<"Reading "<<argv[i]<<endl;
	    ifstream input_file(argv[i]);
	    if(input_file.bad()){
	      cout<<"ERROR: Could not read file \""<<argv[i]<<"\", exiting."<<endl;
	      exit(1);
	    };
	    read_input(input_file,argv[i]);
	    input_file.close();
	  };
	};
      };
    };
  }
  catch(general_exception &the_ex){
    cout<<"*** Error while reading input: "<<the_ex.error_msg<<endl;
    exit(1);
  }
  catch(...){
    cout<<"*** Program ended because of unknown error while reading input"<<endl;
    exit(1);
  };

  if(0 == parameters){
    cout << "*** ERROR: You did not enter the klr parameters"<<endl;
    exit(1);
  };
  if(0 == kernel){
    kernel = new kernel_dot_c();
  };
  if(0 == training_set){
    cout << "*** ERROR: You did not enter the training set"<<endl;
    exit(1);
  };
  if(2 > training_set->size()){
    cout << "*** ERROR: Need at least two examples to learn."<<endl;
    exit(1);
  };

  parameters->do_scale_y = 0;
  parameters->is_pattern=0;
  parameters->is_distribution=0;

  // scale examples
  if(parameters->do_scale){
    training_set->scale(0);
  };

  // copy examples to local vars
  target = training_set->get_ys();
  n = training_set->size();
  C = parameters->realC;
  tol = parameters->convergence_epsilon;
  epsilon = C*parameters->is_zero;
  max_iterations = parameters->max_iterations;
  mu = C*parameters->is_zero;
  N1 = training_set->size_pos();
  N2 = n-N1;
  kernel->init(parameters->kernel_cache,training_set);

  if(training_set->initialised_alpha()){
    // read alphas from training set
    n = training_set->size();
    b = training_set->get_b();
    alpha = training_set->get_alphas();
  }
  else{
    // train the klr
    if(parameters->verbosity > 1){
      cout<<"Computing KLR..."<<endl;
    };
    klr();
    b = -b; // different b in SVM and KLR, set to SVM standard
    
    // copy local vars to training_set
    // save y_i*alpha_i instead of alpha_i
    training_set->put_b(b);
    for(long i=0;i<n;i++){
      if(target[i] < 0){
	alpha[i] = -alpha[i];
      };
      training_set->put_alpha(i,alpha[i]);
    };
    training_set->set_initialised_alpha();

    // save results
    if(parameters->verbosity > 1){
      cout<<"Saving trained KLR to "<<(training_set->get_filename())<<".klr"<<endl;
    };
    char* outname = new char[MAXCHAR];
    strcpy(outname,training_set->get_filename());
    strcat(outname,".klr");
    ofstream output_file(outname,ios::out|ios::trunc);
    output_file.precision(16);
    output_file<<*training_set;
    output_file.close();
    delete []outname;
  };

  if(0 != test_sets){
    example_set_c* next_test;
    long test_no = 0;
    char* outname = new char[MAXCHAR];
    while(test_sets != 0){
      test_no++;
      next_test = test_sets->the_set;
      if(parameters->do_scale){
	next_test->scale(training_set->get_exp(),
			 training_set->get_var(),
			 training_set->get_dim());
      };
      cout<<"Predicting examples from file "<<(next_test->get_filename())<<endl;
      predict(next_test);
      // output to file .pred

      strcpy(outname,next_test->get_filename());
      strcat(outname,".pred");
      ofstream output_file(outname,ios::out);
      output_file<<"@examples"<<endl;
      output_file<<(*next_test);
      output_file.close();	
      test_sets = test_sets->next; // skip delete!
    };
    delete []outname;
  };

  if(kernel) delete kernel;
  if(parameters->verbosity > 1){
    cout << "myklr ended successfully."<<endl;
  };
  return(0);
};
